import numpy as np
import pytest

from autolens.data.array import mask
from autolens.data.array import interpolation
from autolens.model.galaxy import galaxy
from autolens.model.profiles import mass_profiles


@pytest.fixture(name='scheme')
def make_scheme():
    return interpolation.InterpolationScheme(shape=(3, 3), image_coords=np.array([[1.0, 1.0]]), image_pixel_scale=1.0)


@pytest.fixture(name='geometry')
def make_geometry():
    return interpolation.InterpolationGeometry(y_min=-1.0, y_max=1.0, x_min=-1.0, x_max=1.0,
                                               y_pixel_scale=1.0, x_pixel_scale=1.0)


@pytest.fixture(name='galaxy_no_profiles', scope='function')
def make_galaxy_no_profiles():
    return galaxy.Galaxy()


@pytest.fixture(name="galaxy_mass_sis")
def make_galaxy_mass_sis():
    sis = mass_profiles.SphericalIsothermal(einstein_radius=1.0)
    return galaxy.Galaxy(mass_profile=sis)


class TestInterpolationScheme(object):
    class TestConstructor:

        def test__sets_up_attributes_correctly(self):
            image_coords = np.array([[-1.0, -6.0], [-1.0, 0.0], [-4.0, 2.0],
                                     [-0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                     [3.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            interp = interpolation.InterpolationScheme(shape=(3, 3), image_coords=image_coords, image_pixel_scale=1.0)

            assert interp.shape == (3, 3)
            assert interp.pixels == 9
            assert (interp.image_coords == image_coords).all()
            assert interp.geometry.y_min == -6.0
            assert interp.geometry.y_max == 2.0
            assert interp.geometry.x_min == -4.0
            assert interp.geometry.x_max == 3.0
            assert interp.geometry.y_pixel_scale == 1.0
            assert interp.geometry.x_pixel_scale == 1.0
            assert interp.geometry.x_size == 7.0
            assert interp.geometry.y_size == 8.0
            assert interp.geometry.x_start == -4.5
            assert interp.geometry.y_start == -6.5

    class TestNeighbors:

        def test___3x3_grid_neighbors_all_correct(self):
            # |0|1|2|
            # |3|4|5|
            # |6|7|8|

            interp = interpolation.InterpolationScheme(shape=(3, 3), image_coords=np.array([[1.0, 1.0]]),
                                                       image_pixel_scale=1.0)

            assert (interp.bottom_right_neighbors[0] == np.array([1, 3, 4])).all()
            assert (interp.bottom_right_neighbors[1] == np.array([2, 4, 5])).all()
            assert (interp.bottom_right_neighbors[2] == np.array([-1, 5, -1])).all()
            assert (interp.bottom_right_neighbors[3] == np.array([4, 6, 7])).all()
            assert (interp.bottom_right_neighbors[4] == np.array([5, 7, 8])).all()
            assert (interp.bottom_right_neighbors[5] == np.array([-1, 8, -1])).all()
            assert (interp.bottom_right_neighbors[6] == np.array([7, -1, -1])).all()
            assert (interp.bottom_right_neighbors[7] == np.array([8, -1, -1])).all()
            assert (interp.bottom_right_neighbors[8] == np.array([-1, -1, -1])).all()

            assert (interp.bottom_left_neighbors[0] == np.array([-1, -1, 3])).all()
            assert (interp.bottom_left_neighbors[1] == np.array([0, 3, 4])).all()
            assert (interp.bottom_left_neighbors[2] == np.array([1, 4, 5])).all()
            assert (interp.bottom_left_neighbors[3] == np.array([-1, -1, 6])).all()
            assert (interp.bottom_left_neighbors[4] == np.array([3, 6, 7])).all()
            assert (interp.bottom_left_neighbors[5] == np.array([4, 7, 8])).all()
            assert (interp.bottom_left_neighbors[6] == np.array([-1, -1, -1])).all()
            assert (interp.bottom_left_neighbors[7] == np.array([6, -1, -1])).all()
            assert (interp.bottom_left_neighbors[8] == np.array([7, -1, -1])).all()

            assert (interp.top_right_neighbors[0] == np.array([-1, -1, 1])).all()
            assert (interp.top_right_neighbors[1] == np.array([-1, -1, 2])).all()
            assert (interp.top_right_neighbors[2] == np.array([-1, -1, -1])).all()
            assert (interp.top_right_neighbors[3] == np.array([0, 1, 4])).all()
            assert (interp.top_right_neighbors[4] == np.array([1, 2, 5])).all()
            assert (interp.top_right_neighbors[5] == np.array([2, -1, -1])).all()
            assert (interp.top_right_neighbors[6] == np.array([3, 4, 7])).all()
            assert (interp.top_right_neighbors[7] == np.array([4, 5, 8])).all()
            assert (interp.top_right_neighbors[8] == np.array([5, -1, -1])).all()

            assert (interp.top_left_neighbors[0] == np.array([-1, -1, -1])).all()
            assert (interp.top_left_neighbors[1] == np.array([-1, -1, 0])).all()
            assert (interp.top_left_neighbors[2] == np.array([-1, -1, 1])).all()
            assert (interp.top_left_neighbors[3] == np.array([-1, 0, -1])).all()
            assert (interp.top_left_neighbors[4] == np.array([0, 1, 3])).all()
            assert (interp.top_left_neighbors[5] == np.array([1, 2, 4])).all()
            assert (interp.top_left_neighbors[6] == np.array([-1, 3, -1])).all()
            assert (interp.top_left_neighbors[7] == np.array([3, 4, 6])).all()
            assert (interp.top_left_neighbors[8] == np.array([4, 5, 7])).all()

        def test___3x4_grid_neighbors_all_correct(self):
            # |0|1| 2| 3|
            # |4|5| 6| 7|
            # |8|9|10|11|

            interp = interpolation.InterpolationScheme(shape=(3, 4), image_coords=np.array([[1.0, 1.0]]),
                                                       image_pixel_scale=1.0)

            assert (interp.bottom_right_neighbors[0] == np.array([1, 4, 5])).all()
            assert (interp.bottom_right_neighbors[1] == np.array([2, 5, 6])).all()
            assert (interp.bottom_right_neighbors[2] == np.array([3, 6, 7])).all()
            assert (interp.bottom_right_neighbors[3] == np.array([-1, 7, -1])).all()
            assert (interp.bottom_right_neighbors[4] == np.array([5, 8, 9])).all()
            assert (interp.bottom_right_neighbors[5] == np.array([6, 9, 10])).all()
            assert (interp.bottom_right_neighbors[6] == np.array([7, 10, 11])).all()
            assert (interp.bottom_right_neighbors[7] == np.array([-1, 11, -1])).all()
            assert (interp.bottom_right_neighbors[8] == np.array([9, -1, -1])).all()
            assert (interp.bottom_right_neighbors[9] == np.array([10, -1, -1])).all()
            assert (interp.bottom_right_neighbors[10] == np.array([11, -1, -1])).all()
            assert (interp.bottom_right_neighbors[11] == np.array([-1, -1, -1])).all()

            assert (interp.bottom_left_neighbors[0] == np.array([-1, -1, 4])).all()
            assert (interp.bottom_left_neighbors[1] == np.array([0, 4, 5])).all()
            assert (interp.bottom_left_neighbors[2] == np.array([1, 5, 6])).all()
            assert (interp.bottom_left_neighbors[3] == np.array([2, 6, 7])).all()
            assert (interp.bottom_left_neighbors[4] == np.array([-1, -1, 8])).all()
            assert (interp.bottom_left_neighbors[5] == np.array([4, 8, 9])).all()
            assert (interp.bottom_left_neighbors[6] == np.array([5, 9, 10])).all()
            assert (interp.bottom_left_neighbors[7] == np.array([6, 10, 11])).all()
            assert (interp.bottom_left_neighbors[8] == np.array([-1, -1, -1])).all()
            assert (interp.bottom_left_neighbors[9] == np.array([8, -1, -1])).all()
            assert (interp.bottom_left_neighbors[10] == np.array([9, -1, -1])).all()
            assert (interp.bottom_left_neighbors[11] == np.array([10, -1, -1])).all()

            assert (interp.top_right_neighbors[0] == np.array([-1, -1, 1])).all()
            assert (interp.top_right_neighbors[1] == np.array([-1, -1, 2])).all()
            assert (interp.top_right_neighbors[2] == np.array([-1, -1, 3])).all()
            assert (interp.top_right_neighbors[3] == np.array([-1, -1, -1])).all()
            assert (interp.top_right_neighbors[4] == np.array([0, 1, 5])).all()
            assert (interp.top_right_neighbors[5] == np.array([1, 2, 6])).all()
            assert (interp.top_right_neighbors[6] == np.array([2, 3, 7])).all()
            assert (interp.top_right_neighbors[7] == np.array([3, -1, -1])).all()
            assert (interp.top_right_neighbors[8] == np.array([4, 5, 9])).all()
            assert (interp.top_right_neighbors[9] == np.array([5, 6, 10])).all()
            assert (interp.top_right_neighbors[10] == np.array([6, 7, 11])).all()
            assert (interp.top_right_neighbors[11] == np.array([7, -1, -1])).all()

            assert (interp.top_left_neighbors[0] == np.array([-1, -1, -1])).all()
            assert (interp.top_left_neighbors[1] == np.array([-1, -1, 0])).all()
            assert (interp.top_left_neighbors[2] == np.array([-1, -1, 1])).all()
            assert (interp.top_left_neighbors[3] == np.array([-1, -1, 2])).all()
            assert (interp.top_left_neighbors[4] == np.array([-1, 0, -1])).all()
            assert (interp.top_left_neighbors[5] == np.array([0, 1, 4])).all()
            assert (interp.top_left_neighbors[6] == np.array([1, 2, 5])).all()
            assert (interp.top_left_neighbors[7] == np.array([2, 3, 6])).all()
            assert (interp.top_left_neighbors[8] == np.array([-1, 4, -1])).all()
            assert (interp.top_left_neighbors[9] == np.array([4, 5, 8])).all()
            assert (interp.top_left_neighbors[10] == np.array([5, 6, 9])).all()
            assert (interp.top_left_neighbors[11] == np.array([6, 7, 10])).all()

        def test___4x3_grid_neighbors_all_correct(self):
            # |0| 1| 2|
            # |3| 4| 5|
            # |6| 7| 8|
            # |9|10|11|

            interp = interpolation.InterpolationScheme(shape=(4, 3), image_coords=np.array([[1.0, 1.0]]),
                                                       image_pixel_scale=1.0)

            assert (interp.bottom_right_neighbors[0] == np.array([1, 3, 4])).all()
            assert (interp.bottom_right_neighbors[1] == np.array([2, 4, 5])).all()
            assert (interp.bottom_right_neighbors[2] == np.array([-1, 5, -1])).all()
            assert (interp.bottom_right_neighbors[3] == np.array([4, 6, 7])).all()
            assert (interp.bottom_right_neighbors[4] == np.array([5, 7, 8])).all()
            assert (interp.bottom_right_neighbors[5] == np.array([-1, 8, -1])).all()
            assert (interp.bottom_right_neighbors[6] == np.array([7, 9, 10])).all()
            assert (interp.bottom_right_neighbors[7] == np.array([8, 10, 11])).all()
            assert (interp.bottom_right_neighbors[8] == np.array([-1, 11, -1])).all()
            assert (interp.bottom_right_neighbors[9] == np.array([10, -1, -1])).all()
            assert (interp.bottom_right_neighbors[10] == np.array([11, -1, -1])).all()
            assert (interp.bottom_right_neighbors[11] == np.array([-1, -1, -1])).all()

            assert (interp.bottom_left_neighbors[0] == np.array([-1, -1, 3])).all()
            assert (interp.bottom_left_neighbors[1] == np.array([0, 3, 4])).all()
            assert (interp.bottom_left_neighbors[2] == np.array([1, 4, 5])).all()
            assert (interp.bottom_left_neighbors[3] == np.array([-1, -1, 6])).all()
            assert (interp.bottom_left_neighbors[4] == np.array([3, 6, 7])).all()
            assert (interp.bottom_left_neighbors[5] == np.array([4, 7, 8])).all()
            assert (interp.bottom_left_neighbors[6] == np.array([-1, -1, 9])).all()
            assert (interp.bottom_left_neighbors[7] == np.array([6, 9, 10])).all()
            assert (interp.bottom_left_neighbors[8] == np.array([7, 10, 11])).all()
            assert (interp.bottom_left_neighbors[9] == np.array([-1, -1, -1])).all()
            assert (interp.bottom_left_neighbors[10] == np.array([9, -1, -1])).all()
            assert (interp.bottom_left_neighbors[11] == np.array([10, -1, -1])).all()

            assert (interp.top_right_neighbors[0] == np.array([-1, -1, 1])).all()
            assert (interp.top_right_neighbors[1] == np.array([-1, -1, 2])).all()
            assert (interp.top_right_neighbors[2] == np.array([-1, -1, -1])).all()
            assert (interp.top_right_neighbors[3] == np.array([0, 1, 4])).all()
            assert (interp.top_right_neighbors[4] == np.array([1, 2, 5])).all()
            assert (interp.top_right_neighbors[5] == np.array([2, -1, -1])).all()
            assert (interp.top_right_neighbors[6] == np.array([3, 4, 7])).all()
            assert (interp.top_right_neighbors[7] == np.array([4, 5, 8])).all()
            assert (interp.top_right_neighbors[8] == np.array([5, -1, -1])).all()
            assert (interp.top_right_neighbors[9] == np.array([6, 7, 10])).all()
            assert (interp.top_right_neighbors[10] == np.array([7, 8, 11])).all()
            assert (interp.top_right_neighbors[11] == np.array([8, -1, -1])).all()

            assert (interp.top_left_neighbors[0] == np.array([-1, -1, -1])).all()
            assert (interp.top_left_neighbors[1] == np.array([-1, -1, 0])).all()
            assert (interp.top_left_neighbors[2] == np.array([-1, -1, 1])).all()
            assert (interp.top_left_neighbors[3] == np.array([-1, 0, -1])).all()
            assert (interp.top_left_neighbors[4] == np.array([0, 1, 3])).all()
            assert (interp.top_left_neighbors[5] == np.array([1, 2, 4])).all()
            assert (interp.top_left_neighbors[6] == np.array([-1, 3, -1])).all()
            assert (interp.top_left_neighbors[7] == np.array([3, 4, 6])).all()
            assert (interp.top_left_neighbors[8] == np.array([4, 5, 7])).all()
            assert (interp.top_left_neighbors[9] == np.array([-1, 6, -1])).all()
            assert (interp.top_left_neighbors[10] == np.array([6, 7, 9])).all()
            assert (interp.top_left_neighbors[11] == np.array([7, 8, 10])).all()

        def test___4x4_grid_neighbors_all_correct(self):
            # | 0| 1| 2| 3|
            # | 4| 5| 6| 7|
            # | 8| 9|10|11|
            # |12|13|14|15|

            interp = interpolation.InterpolationScheme(shape=(4, 4), image_coords=np.array([[1.0, 1.0]]),
                                                       image_pixel_scale=1.0)

            assert (interp.bottom_right_neighbors[0] == np.array([1, 4, 5])).all()
            assert (interp.bottom_right_neighbors[1] == np.array([2, 5, 6])).all()
            assert (interp.bottom_right_neighbors[2] == np.array([3, 6, 7])).all()
            assert (interp.bottom_right_neighbors[3] == np.array([-1, 7, -1])).all()
            assert (interp.bottom_right_neighbors[4] == np.array([5, 8, 9])).all()
            assert (interp.bottom_right_neighbors[5] == np.array([6, 9, 10])).all()
            assert (interp.bottom_right_neighbors[6] == np.array([7, 10, 11])).all()
            assert (interp.bottom_right_neighbors[7] == np.array([-1, 11, -1])).all()
            assert (interp.bottom_right_neighbors[8] == np.array([9, 12, 13])).all()
            assert (interp.bottom_right_neighbors[9] == np.array([10, 13, 14])).all()
            assert (interp.bottom_right_neighbors[10] == np.array([11, 14, 15])).all()
            assert (interp.bottom_right_neighbors[11] == np.array([-1, 15, -1])).all()
            assert (interp.bottom_right_neighbors[12] == np.array([13, -1, -1])).all()
            assert (interp.bottom_right_neighbors[13] == np.array([14, -1, -1])).all()
            assert (interp.bottom_right_neighbors[14] == np.array([15, -1, -1])).all()
            assert (interp.bottom_right_neighbors[15] == np.array([-1, -1, -1])).all()

            assert (interp.bottom_left_neighbors[0] == np.array([-1, -1, 4])).all()
            assert (interp.bottom_left_neighbors[1] == np.array([0, 4, 5])).all()
            assert (interp.bottom_left_neighbors[2] == np.array([1, 5, 6])).all()
            assert (interp.bottom_left_neighbors[3] == np.array([2, 6, 7])).all()
            assert (interp.bottom_left_neighbors[4] == np.array([-1, -1, 8])).all()
            assert (interp.bottom_left_neighbors[5] == np.array([4, 8, 9])).all()
            assert (interp.bottom_left_neighbors[6] == np.array([5, 9, 10])).all()
            assert (interp.bottom_left_neighbors[7] == np.array([6, 10, 11])).all()
            assert (interp.bottom_left_neighbors[8] == np.array([-1, -1, 12])).all()
            assert (interp.bottom_left_neighbors[9] == np.array([8, 12, 13])).all()
            assert (interp.bottom_left_neighbors[10] == np.array([9, 13, 14])).all()
            assert (interp.bottom_left_neighbors[11] == np.array([10, 14, 15])).all()
            assert (interp.bottom_left_neighbors[12] == np.array([-1, -1, -1])).all()
            assert (interp.bottom_left_neighbors[13] == np.array([12, -1, -1])).all()
            assert (interp.bottom_left_neighbors[14] == np.array([13, -1, -1])).all()
            assert (interp.bottom_left_neighbors[15] == np.array([14, -1, -1])).all()

            assert (interp.top_right_neighbors[0] == np.array([-1, -1, 1])).all()
            assert (interp.top_right_neighbors[1] == np.array([-1, -1, 2])).all()
            assert (interp.top_right_neighbors[2] == np.array([-1, -1, 3])).all()
            assert (interp.top_right_neighbors[3] == np.array([-1, -1, -1])).all()
            assert (interp.top_right_neighbors[4] == np.array([0, 1, 5])).all()
            assert (interp.top_right_neighbors[5] == np.array([1, 2, 6])).all()
            assert (interp.top_right_neighbors[6] == np.array([2, 3, 7])).all()
            assert (interp.top_right_neighbors[7] == np.array([3, -1, -1])).all()
            assert (interp.top_right_neighbors[8] == np.array([4, 5, 9])).all()
            assert (interp.top_right_neighbors[9] == np.array([5, 6, 10])).all()
            assert (interp.top_right_neighbors[10] == np.array([6, 7, 11])).all()
            assert (interp.top_right_neighbors[11] == np.array([7, -1, -1])).all()
            assert (interp.top_right_neighbors[12] == np.array([8, 9, 13])).all()
            assert (interp.top_right_neighbors[13] == np.array([9, 10, 14])).all()
            assert (interp.top_right_neighbors[14] == np.array([10, 11, 15])).all()
            assert (interp.top_right_neighbors[15] == np.array([11, -1, -1])).all()

            assert (interp.top_left_neighbors[0] == np.array([-1, -1, -1])).all()
            assert (interp.top_left_neighbors[1] == np.array([-1, -1, 0])).all()
            assert (interp.top_left_neighbors[2] == np.array([-1, -1, 1])).all()
            assert (interp.top_left_neighbors[3] == np.array([-1, -1, 2])).all()
            assert (interp.top_left_neighbors[4] == np.array([-1, 0, -1])).all()
            assert (interp.top_left_neighbors[5] == np.array([0, 1, 4])).all()
            assert (interp.top_left_neighbors[6] == np.array([1, 2, 5])).all()
            assert (interp.top_left_neighbors[7] == np.array([2, 3, 6])).all()
            assert (interp.top_left_neighbors[8] == np.array([-1, 4, -1])).all()
            assert (interp.top_left_neighbors[9] == np.array([4, 5, 8])).all()
            assert (interp.top_left_neighbors[10] == np.array([5, 6, 9])).all()
            assert (interp.top_left_neighbors[11] == np.array([6, 7, 10])).all()
            assert (interp.top_left_neighbors[12] == np.array([-1, 8, -1])).all()
            assert (interp.top_left_neighbors[13] == np.array([8, 9, 12])).all()
            assert (interp.top_left_neighbors[14] == np.array([9, 10, 13])).all()
            assert (interp.top_left_neighbors[15] == np.array([10, 11, 14])).all()

    class TestFromMask:

        def test__passes_mask_pixel_scale(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 3))

            assert interp.image_pixel_scale == msk.pixel_scale

        def test__3x3_mask_with_1_pixel__3x3_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.0, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.0, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.0, 1.0]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([0.0, -1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, 1.0]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([1.0, -1.0]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([1.0, 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([1.0, 1.0]), 1e-4)

        def test__same_as_above__change_pixel_scale(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=2.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-2.0, -2.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-2.0, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-2.0, 2.0]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([0.0, -2.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, 2.0]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([2.0, -2.0]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([2.0, 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([2.0, 2.0]), 1e-4)

        def test__3x3_mask_with_1_pixel__4x4_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(4, 4))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.0, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.0, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.0, (1. / 3.)]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([-1.0, 1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([-(1. / 3.), -1.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([-(1. / 3.), -(1. / 3.)]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([-(1. / 3.), (1. / 3.)]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([-(1. / 3.), 1.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([(1. / 3.), -1.0]), 1e-4)
            assert interp.image_coords[9] == pytest.approx(np.array([(1. / 3.), -(1. / 3.)]), 1e-4)
            assert interp.image_coords[10] == pytest.approx(np.array([(1. / 3.), (1. / 3.)]), 1e-4)
            assert interp.image_coords[11] == pytest.approx(np.array([(1. / 3.), 1.0]), 1e-4)
            assert interp.image_coords[12] == pytest.approx(np.array([1.0, -1.0]), 1e-4)
            assert interp.image_coords[13] == pytest.approx(np.array([1.0, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[14] == pytest.approx(np.array([1.0, (1. / 3.)]), 1e-4)
            assert interp.image_coords[15] == pytest.approx(np.array([1.0, 1.0]), 1e-4)

        def test__3x3_mask_with_1_pixel__3x4_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 4))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.0, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.0, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.0, (1. / 3.)]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([-1.0, 1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, -1.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([0.0, (1. / 3.)]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([0.0, 1.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([1.0, -1.0]), 1e-4)
            assert interp.image_coords[9] == pytest.approx(np.array([1.0, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[10] == pytest.approx(np.array([1.0, (1. / 3.)]), 1e-4)
            assert interp.image_coords[11] == pytest.approx(np.array([1.0, 1.0]), 1e-4)

        def test__3x3_mask_with_1_pixel__4x3_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(4, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.0, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.0, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.0, 1.0]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([-(1. / 3.), -1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([-(1. / 3.), 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([-(1. / 3.), 1.0]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([(1. / 3.), -1.0]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([(1. / 3.), 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([(1. / 3.), 1.0]), 1e-4)
            assert interp.image_coords[9] == pytest.approx(np.array([1.0, -1.0]), 1e-4)
            assert interp.image_coords[10] == pytest.approx(np.array([1.0, 0.0]), 1e-4)
            assert interp.image_coords[11] == pytest.approx(np.array([1.0, 1.0]), 1e-4)

        def test__4x4_mask_with_4_pixels__3x3_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True, True],
                            [True, False, False, True],
                            [True, False, False, True],
                            [True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.5, -1.5]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.5, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.5, 1.5]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([0.0, -1.5]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, 1.5]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([1.5, -1.5]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([1.5, 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([1.5, 1.5]), 1e-4)

        def test__3x4_mask_with_2_pixels__3x3_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True, True],
                            [True, False, False, True],
                            [True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.5, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.5, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.5, 1.0]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([0.0, -1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, 1.0]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([1.5, -1.0]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([1.5, 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([1.5, 1.0]), 1e-4)

        def test__4x3_mask_with_4_pixels__3x3_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True],
                            [True, False, True],
                            [True, False, True],
                            [True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.0, -1.5]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.0, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.0, 1.5]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([0.0, -1.5]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, 1.5]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([1.0, -1.5]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([1.0, 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([1.0, 1.5]), 1e-4)

        def test__3x4_mask_with_2_pixels__3x4_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True, True],
                            [True, False, False, True],
                            [True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(3, 4))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.5, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.5, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.5, (1. / 3.)]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([-1.5, 1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([0.0, -1.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([0.0, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([0.0, (1. / 3.)]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([0.0, 1.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([1.5, -1.0]), 1e-4)
            assert interp.image_coords[9] == pytest.approx(np.array([1.5, -(1. / 3.)]), 1e-4)
            assert interp.image_coords[10] == pytest.approx(np.array([1.5, (1. / 3.)]), 1e-4)
            assert interp.image_coords[11] == pytest.approx(np.array([1.5, 1.0]), 1e-4)

        def test__4x3_mask_with_2_pixels__4x3_interp_grid__image_coords_extend_beyond_mask(self):
            msk = np.array([[True, True, True, True],
                            [True, False, False, True],
                            [True, True, True, True]])

            msk = mask.Mask(msk, pixel_scale=1.0)

            interp = interpolation.InterpolationScheme.from_mask(mask=msk, shape=(4, 3))

            assert interp.image_coords[0] == pytest.approx(np.array([-1.5, -1.0]), 1e-4)
            assert interp.image_coords[1] == pytest.approx(np.array([-1.5, 0.0]), 1e-4)
            assert interp.image_coords[2] == pytest.approx(np.array([-1.5, 1.0]), 1e-4)
            assert interp.image_coords[3] == pytest.approx(np.array([-0.5, -1.0]), 1e-4)
            assert interp.image_coords[4] == pytest.approx(np.array([-0.5, 0.0]), 1e-4)
            assert interp.image_coords[5] == pytest.approx(np.array([-0.5, 1.0]), 1e-4)
            assert interp.image_coords[6] == pytest.approx(np.array([0.5, -1.0]), 1e-4)
            assert interp.image_coords[7] == pytest.approx(np.array([0.5, 0.0]), 1e-4)
            assert interp.image_coords[8] == pytest.approx(np.array([0.5, 1.0]), 1e-4)
            assert interp.image_coords[9] == pytest.approx(np.array([1.5, -1.0]), 1e-4)
            assert interp.image_coords[10] == pytest.approx(np.array([1.5, 0.0]), 1e-4)
            assert interp.image_coords[11] == pytest.approx(np.array([1.5, 1.0]), 1e-4)

    class TestInterpolationCoordinatesFromSizes:

        def test__x_and_y_are_same_sizes__scales_are_1__new_coordinates_are_image_coordinates(self):
            image_coords = np.array([[-1.0, -6.0], [-1.0, 0.0], [-4.0, 2.0],
                                     [-0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                     [3.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            interp = interpolation.InterpolationScheme(shape=(3, 3), image_coords=image_coords, image_pixel_scale=1.0)

            interp_coords = interp.interpolation_coordinates_from_sizes(new_x_size=7.0, new_y_size=8.0)

            assert (interp_coords == image_coords).all()
            assert interp_coords.geometry.y_min == -6.0
            assert interp_coords.geometry.y_max == 2.0
            assert interp_coords.geometry.x_min == -4.0
            assert interp_coords.geometry.x_max == 3.0
            assert interp_coords.geometry.y_size == 8.0
            assert interp_coords.geometry.x_size == 7.0
            assert interp_coords.geometry.y_pixel_scale == 1.0
            assert interp_coords.geometry.x_pixel_scale == 1.0
            assert interp_coords.geometry.x_start == -4.5
            assert interp_coords.geometry.y_start == -6.5
            assert interp_coords.scheme == interp

        def test__same_as_above_but_x_dimension_halves(self):
            image_coords = np.array([[-1.0, -6.0], [-1.0, 0.0], [-4.0, 2.0],
                                     [-0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                     [3.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            interp = interpolation.InterpolationScheme(shape=(3, 3), image_coords=image_coords, image_pixel_scale=1.0)

            interp_coords = interp.interpolation_coordinates_from_sizes(new_x_size=3.5, new_y_size=8.0)

            assert (interp_coords == np.array([[-0.5, -6.0], [-0.5, 0.0], [-2.0, 2.0],
                                               [-0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                               [1.5, -1.0], [0.5, 0.0], [0.5, 1.0]])).all()
            assert interp_coords.geometry.y_min == -6.0
            assert interp_coords.geometry.y_max == 2.0
            assert interp_coords.geometry.x_min == -2.0
            assert interp_coords.geometry.x_max == 1.5
            assert interp_coords.geometry.y_size == 8.0
            assert interp_coords.geometry.x_size == 3.5
            assert interp_coords.geometry.y_pixel_scale == 1.0
            assert interp_coords.geometry.x_pixel_scale == 0.5
            assert interp_coords.geometry.x_start == -2.25
            assert interp_coords.geometry.y_start == -6.5
            assert (interp_coords.scheme == interp)

        def test__same_as_above_but_y_dimension_halves(self):
            image_coords = np.array([[-1.0, -6.0], [-1.0, 0.0], [-4.0, 2.0],
                                     [-0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                     [3.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            interp = interpolation.InterpolationScheme(shape=(3, 3), image_coords=image_coords, image_pixel_scale=1.0)

            interp_coords = interp.interpolation_coordinates_from_sizes(new_x_size=7.0, new_y_size=4.0)

            assert (interp_coords == np.array([[-1.0, -3.0], [-1.0, 0.0], [-4.0, 1.0],
                                               [-0.0, -0.5], [0.0, 0.0], [0.0, 0.5],
                                               [3.0, -0.5], [1.0, 0.0], [1.0, 0.5]])).all()
            assert interp_coords.geometry.y_min == -3.0
            assert interp_coords.geometry.y_max == 1.0
            assert interp_coords.geometry.x_min == -4.0
            assert interp_coords.geometry.x_max == 3.0
            assert interp_coords.geometry.y_size == 4.0
            assert interp_coords.geometry.x_size == 7.0
            assert interp_coords.geometry.y_pixel_scale == 0.5
            assert interp_coords.geometry.x_pixel_scale == 1.0
            assert interp_coords.geometry.y_start == -3.25
            assert interp_coords.geometry.x_start == -4.5
            assert (interp_coords.scheme == interp)

        def test__same_as_above_triple_both_dimensions(self):
            image_coords = np.array([[-1.0, -6.0], [-1.0, 0.0], [-4.0, 2.0],
                                     [-0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                     [3.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            interp = interpolation.InterpolationScheme(shape=(3, 3), image_coords=image_coords, image_pixel_scale=1.0)

            interp_coords = interp.interpolation_coordinates_from_sizes(new_x_size=21.0, new_y_size=24.0)

            assert (interp_coords == np.array([[-3.0, -18.0], [-3.0, 0.0], [-12.0, 6.0],
                                               [-0.0, -3.0], [0.0, 0.0], [0.0, 3.0],
                                               [9.0, -3.0], [3.0, 0.0], [3.0, 3.0]])).all()
            assert interp_coords.geometry.y_min == -18.0
            assert interp_coords.geometry.y_max == 6.0
            assert interp_coords.geometry.x_min == -12.0
            assert interp_coords.geometry.x_max == 9.0
            assert interp_coords.geometry.y_size == 24.0
            assert interp_coords.geometry.x_size == 21.0
            assert interp_coords.geometry.y_pixel_scale == 3.0
            assert interp_coords.geometry.x_pixel_scale == 3.0
            assert interp_coords.geometry.x_start == -13.5
            assert interp_coords.geometry.y_start == -19.5
            assert (interp_coords.scheme == interp)


class TestInterpolationCoordinates(object):
    class TestConstructor:

        def test__sets_up_coordinates_correctly(self):
            geometry = interpolation.InterpolationGeometry(y_min=-1.0, y_max=1.0, x_min=-1.0, x_max=1.0,
                                                           y_pixel_scale=1.0, x_pixel_scale=1.0)

            scheme = interpolation.InterpolationScheme(shape=(3, 3), image_coords=np.array([[1.0, 1.0]]),
                                                       image_pixel_scale=1.0)

            interp_coords = interpolation.InterpolationCoordinates(array=np.array([[0.0, 0.0], [1.0, 1.0], [2.0, 2.0]]),
                                                                   geometry=geometry, scheme=scheme)

            assert (interp_coords == np.array([[0.0, 0.0], [1.0, 1.0], [2.0, 2.0]])).all()
            assert interp_coords.geometry.y_min == -1.0
            assert interp_coords.geometry.y_max == 1.0
            assert interp_coords.geometry.x_min == -1.0
            assert interp_coords.geometry.x_max == 1.0
            assert interp_coords.geometry.y_size == 2.0
            assert interp_coords.geometry.x_size == 2.0
            assert interp_coords.geometry.y_pixel_scale == 1.0
            assert interp_coords.geometry.x_pixel_scale == 1.0
            assert interp_coords.geometry.y_start == -1.5
            assert interp_coords.geometry.x_start == -1.5
            assert interp_coords.scheme.shape == (3, 3)
            assert (interp_coords.scheme.image_coords == np.array([[1.0, 1.0]])).all()

    class TestInterpolationDeflectionsFromCoordinates:

        def test__galaxy_has_no_mass_profile__deflection_coordinates_are_0s(self, galaxy_no_profiles, geometry, scheme):
            interp_coords = interpolation.InterpolationCoordinates(array=np.array([[1.0, 1.0], [1.0, 0.0]]),
                                                                   geometry=geometry, scheme=scheme)

            interp_defls = interp_coords.interpolation_deflections_from_coordinates_and_galaxies(
                galaxies=[galaxy_no_profiles])

            assert (interp_defls == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
            assert (interp_defls.interp_coords == np.array([[1.0, 1.0], [1.0, 0.0]])).all()
            assert interp_defls.geometry == geometry
            assert interp_defls.scheme == scheme

        def test__galaxy_has_sis_mass_profile__deflection_coordinates_are_correct(self, galaxy_mass_sis, geometry,
                                                                                  scheme):
            interp_coords = interpolation.InterpolationCoordinates(array=np.array([[1.0, 1.0], [1.0, 0.0]]),
                                                                   geometry=geometry, scheme=scheme)

            interp_defls = interp_coords.interpolation_deflections_from_coordinates_and_galaxies(
                galaxies=[galaxy_mass_sis])

            assert interp_defls[0] == pytest.approx(np.array([0.707, 0.707]), 1e-3)
            assert interp_defls[1] == pytest.approx(np.array([1.0, 0.0]), 1e-3)
            assert (interp_defls.interp_coords == np.array([[1.0, 1.0], [1.0, 0.0]])).all()
            assert interp_defls.geometry == geometry
            assert interp_defls.scheme == scheme

        def test__x3_galaxies_with_sis_mass_profile__deflection_coordinates_are_correct(self, galaxy_mass_sis, geometry,
                                                                                        scheme):
            interp_coords = interpolation.InterpolationCoordinates(array=np.array([[1.0, 1.0], [1.0, 0.0]]),
                                                                   geometry=geometry, scheme=scheme)

            interp_defls = interp_coords.interpolation_deflections_from_coordinates_and_galaxies(
                galaxies=[galaxy_mass_sis, galaxy_mass_sis, galaxy_mass_sis])

            assert interp_defls[0] == pytest.approx(np.array([3.0 * 0.707, 3.0 * 0.707]), 1e-3)
            assert interp_defls[1] == pytest.approx(np.array([3.0, 0.0]), 1e-3)
            assert (interp_defls.interp_coords == np.array([[1.0, 1.0], [1.0, 0.0]])).all()
            assert interp_defls.geometry == geometry
            assert interp_defls.scheme == scheme


class TestInterpolationDeflections(object):
    class TestConstructor:

        def test__sets_up_attributes_correctly(self, geometry, scheme):
            interp_defls = interpolation.InterpolationDeflections(array=np.array([[1.0, 1.0]]),
                                                                  coords=np.array([[2.0, 2.0]]), geometry=geometry,
                                                                  scheme=scheme)

            assert (interp_defls == np.array([[1.0, 1.0]])).all()
            assert (interp_defls.interp_coords == np.array([[2.0, 2.0]])).all()
            assert interp_defls.geometry == geometry
            assert interp_defls.scheme == scheme

    class TestGridToInterp:

        def test__3x3_interp_scheme__coordinate_on_centre_of_each_pixel__grid_retuns_correct_index(self):
            # |0|1|2|
            # |3|4|5|
            # |6|7|8|

            # interp defls arn't ussed for grid to interp

            interp_defls = np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

            interp_coords = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                      [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                      [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            geometry = interpolation.InterpolationGeometry(y_min=-1.0, y_max=1.0, x_min=-1.0, x_max=1.0,
                                                           y_pixel_scale=1.0, x_pixel_scale=1.0)
            scheme = interpolation.InterpolationScheme(shape=(3, 3), image_coords=interp_coords, image_pixel_scale=1.0)

            interp_defls = interpolation.InterpolationDeflections(array=interp_defls, coords=interp_coords,
                                                                  geometry=geometry, scheme=scheme)

            grid_to_interp = interp_defls.grid_to_interp_from_grid(
                grid=np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                               [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                               [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]]))

            assert (grid_to_interp[0] == np.array([0])).all()
            assert (grid_to_interp[1] == np.array([1])).all()
            assert (grid_to_interp[2] == np.array([2])).all()
            assert (grid_to_interp[3] == np.array([3])).all()
            assert (grid_to_interp[4] == np.array([4])).all()
            assert (grid_to_interp[5] == np.array([5])).all()
            assert (grid_to_interp[6] == np.array([6])).all()
            assert (grid_to_interp[7] == np.array([7])).all()
            assert (grid_to_interp[8] == np.array([8])).all()

        def test__3x3_interp_scheme__coordinates_near_edges_of_pixels__different_ordering(self):
            # |0|1|2|
            # |3|4|5|
            # |6|7|8|

            # interp defls arn't ussed for bilinear interpolation weights

            interp_defls = np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0],
                                     [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

            interp_coords = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
                                      [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                      [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]])

            geometry = interpolation.InterpolationGeometry(y_min=-1.0, y_max=1.0, x_min=-1.0, x_max=1.0,
                                                           y_pixel_scale=1.0, x_pixel_scale=1.0)
            scheme = interpolation.InterpolationScheme(shape=(3, 3), image_coords=interp_coords, image_pixel_scale=1.0)

            interp_defls = interpolation.InterpolationDeflections(array=interp_defls, coords=interp_coords,
                                                                  geometry=geometry, scheme=scheme)

            grid_to_interp = interp_defls.grid_to_interp_from_grid(
                grid=np.array([[0.501, 0.501], [-1.49, 0.49], [-0.499, 0.499],
                               [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                               [1.0, -1.0], [1.0, 0.0], [1.0, 1.0]]))

            assert (grid_to_interp[0] == np.array([8])).all()
            assert (grid_to_interp[1] == np.array([1])).all()
            assert (grid_to_interp[2] == np.array([4])).all()
            assert (grid_to_interp[3] == np.array([3])).all()
            assert (grid_to_interp[4] == np.array([4])).all()
            assert (grid_to_interp[5] == np.array([5])).all()
            assert (grid_to_interp[6] == np.array([6])).all()
            assert (grid_to_interp[7] == np.array([7])).all()
            assert (grid_to_interp[8] == np.array([8])).all()

    # class TestInterpolateValues:
    #
    #     class TestInterpolateTopLeft:
    #
    #         def test__1_coordinate__at_top_left_of_pixel__interpolates_to_average_of_4_values(self):
    #
    #             # |0|1|2|
    #             # |3|4|5|
    #             # |6|7|8|
    #
    #             # interp defls arn't ussed for bilinear interpolation weights
    #
    #             interp_defls = np.array([[ 1.0,  1.0], [ 1.0, 1.0], [ 0.0, 0.0],
    #                                      [ 1.0,  1.0], [ 1.0, 1.0], [ 0.0, 0.0],
    #                                      [ 0.0,  0.0], [ 0.0, 0.0], [ 0.0, 0.0]])
    #
    #             interp_coords = np.array([[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0],
    #                                       [ 0.0, -1.0], [ 0.0, 0.0], [ 0.0, 1.0],
    #                                       [ 1.0, -1.0], [ 1.0, 0.0], [ 1.0, 1.0]])
    #
    #             geometry = interpolation.InterpolationGeometry(y_min=-1.0, y_max=1.0, x_min=-1.0, x_max=1.0,
    #                                                            y_pixel_scale=1.0, x_pixel_scale=1.0)
    #             scheme = interpolation.InterpolationScheme(shape=(3, 3), datas_=interp_coords,
    #                                                        image_pixel_scale=1.0)
    #
    #             interp_defls = interpolation.InterpolationDeflections(array=interp_defls, grid_stacks=interp_coords,
    #                                                                   geometry=geometry, scheme=scheme)
    #
    #             interpolated = interp_defls.interpolate_values_from_grid(grid=np.array([[-0.499, -0.499]]))
    #
    #             assert interpolated[0,0] == 1.0
    #             assert interpolated[0,0] == 1.0



import numpy as np
import pytest

from autolens.data import ccd as im, convolution
from autolens.data.array.util import grid_util
from autolens.data.array import scaled_array
from autolens.data.array import mask as msk
from autolens.lens.stack import lens_data_stack as lis
from autolens.model.inversion import convolution as inversion_convolution


@pytest.fixture(name='ccd_0')
def make_image_0():

    image = scaled_array.ScaledSquarePixelArray(array=np.ones((4, 4)), pixel_scale=3.0)
    psf = im.PSF(array=np.ones((3, 3)), pixel_scale=3.0, renormalize=False)
    noise_map = im.NoiseMap(array=2.0*np.ones((4,4)), pixel_scale=3.0)
    background_noise_map = im.NoiseMap(array=3.0*np.ones((4,4)), pixel_scale=3.0)
    poisson_noise_map = im.PoissonNoiseMap(array=4.0*np.ones((4,4)), pixel_scale=3.0)
    exposure_time_map = im.ExposureTimeMap(array=5.0 * np.ones((4, 4)), pixel_scale=3.0)
    background_sky_map = scaled_array.ScaledSquarePixelArray(array=6.0 * np.ones((4, 4)), pixel_scale=3.0)

    return im.CCDData(image=image, pixel_scale=3.0, psf=psf, noise_map=noise_map,
                      background_noise_map=background_noise_map, poisson_noise_map=poisson_noise_map,
                      exposure_time_map=exposure_time_map, background_sky_map=background_sky_map)

@pytest.fixture(name='ccd_1')
def make_image_1():

    image = scaled_array.ScaledSquarePixelArray(array=10.0*np.ones((4, 4)), pixel_scale=6.0)
    psf = im.PSF(array=11.0*np.ones((3, 3)), pixel_scale=6.0, renormalize=False)
    noise_map = im.NoiseMap(array=12.0*np.ones((4,4)), pixel_scale=6.0)
    background_noise_map = im.NoiseMap(array=13.0*np.ones((4,4)), pixel_scale=6.0)
    poisson_noise_map = im.PoissonNoiseMap(array=14.0*np.ones((4,4)), pixel_scale=6.0)
    exposure_time_map = im.ExposureTimeMap(array=15.0 * np.ones((4, 4)), pixel_scale=6.0)
    background_sky_map = scaled_array.ScaledSquarePixelArray(array=16.0 * np.ones((4, 4)), pixel_scale=6.0)

    return im.CCDData(image=image, pixel_scale=6.0, psf=psf, noise_map=noise_map,
                      background_noise_map=background_noise_map, poisson_noise_map=poisson_noise_map,
                      exposure_time_map=exposure_time_map, background_sky_map=background_sky_map)

@pytest.fixture(name="mask_0")
def make_mask_0():
    return msk.Mask(np.array([[True, True, True, True],
                              [True, False, False, True],
                              [True, False, False, True],
                              [True, True, True, True]]), pixel_scale=3.0)

@pytest.fixture(name="mask_1")
def make_mask_1():
    return msk.Mask(np.array([[True, True, True, True],
                              [True, False, False, True],
                              [True, False, False, True],
                              [True, True, True, True]]), pixel_scale=6.0)

@pytest.fixture(name="lens_data_stack")
def make_lens_image(ccd_0, ccd_1, mask_0, mask_1):
    return lis.LensDataStack(ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1])


class TestLensData(object):

    def test_attributes(self, ccd_0, ccd_1, lens_data_stack):

        assert lens_data_stack.pixel_scales[0] == ccd_0.pixel_scale
        assert lens_data_stack.pixel_scales[0] == 3.0
        assert (lens_data_stack.images[0] == ccd_0.image).all()
        assert (lens_data_stack.images[0] == np.ones((4,4))).all()
        assert (lens_data_stack.psfs[0] == ccd_0.psf).all()
        assert (lens_data_stack.psfs[0] == np.ones((3,3))).all()
        assert (lens_data_stack.noise_maps[0] == ccd_0.noise_map).all()
        assert (lens_data_stack.noise_maps[0] == 2.0*np.ones((4,4))).all()

        assert lens_data_stack.pixel_scales[1] == ccd_1.pixel_scale
        assert lens_data_stack.pixel_scales[1] == 6.0
        assert (lens_data_stack.images[1] == ccd_1.image).all()
        assert (lens_data_stack.images[1] == 10.0*np.ones((4,4))).all()
        assert (lens_data_stack.psfs[1] == ccd_1.psf).all()
        assert (lens_data_stack.psfs[1] == 11.0*np.ones((3,3))).all()
        assert (lens_data_stack.noise_maps[1] == ccd_1.noise_map).all()
        assert (lens_data_stack.noise_maps[1] == 12.0*np.ones((4,4))).all()

    def test_masking(self, lens_data_stack):

        assert (lens_data_stack.images_1d[0] == np.ones(4)).all()
        assert (lens_data_stack.noise_maps_1d[0] == 2.0*np.ones(4)).all()
        assert (lens_data_stack.masks_1d[0] == np.array([False, False, False, False])).all()

        assert (lens_data_stack.images_1d[1] == 10.0*np.ones(4)).all()
        assert (lens_data_stack.noise_maps_1d[1] == 12.0*np.ones(4)).all()
        assert (lens_data_stack.masks_1d[1] == np.array([False, False, False, False])).all()

    def test_grids(self, lens_data_stack):

        assert (lens_data_stack.grid_stacks[0].regular ==
                np.array([[1.5, -1.5], [1.5, 1.5], [-1.5, -1.5], [-1.5, 1.5]])).all()

        assert (lens_data_stack.grid_stacks[0].sub ==
                np.array([[2.0, -2.0], [2.0, -1.0], [1.0, -2.0], [1.0, -1.0],
                          [2.0, 1.0], [2.0, 2.0], [1.0, 1.0], [1.0, 2.0],
                          [-1.0, -2.0], [-1.0, -1.0], [-2.0, -2.0], [-2.0, -1.0],
                          [-1.0, 1.0], [-1.0, 2.0], [-2.0, 1.0], [-2.0, 2.0]])).all()

        assert (lens_data_stack.grid_stacks[0].blurring ==
                np.array([[4.5, -4.5], [4.5, -1.5], [4.5, 1.5], [4.5, 4.5],
                          [1.5, -4.5], [1.5, 4.5], [-1.5, -4.5], [-1.5, 4.5],
                          [-4.5, -4.5], [-4.5, -1.5], [-4.5, 1.5], [-4.5, 4.5]])).all()

        # Pixel scale is doubled, thus the grids of coordinates are doubled.

        assert (lens_data_stack.grid_stacks[1].regular ==
                2.0*np.array([[1.5, -1.5], [1.5, 1.5], [-1.5, -1.5], [-1.5, 1.5]])).all()

        assert (lens_data_stack.grid_stacks[1].sub ==
                2.0*np.array([[2.0, -2.0], [2.0, -1.0], [1.0, -2.0], [1.0, -1.0],
                              [2.0, 1.0], [2.0, 2.0], [1.0, 1.0], [1.0, 2.0],
                              [-1.0, -2.0], [-1.0, -1.0], [-2.0, -2.0], [-2.0, -1.0],
                              [-1.0, 1.0], [-1.0, 2.0], [-2.0, 1.0], [-2.0, 2.0]])).all()

        assert (lens_data_stack.grid_stacks[1].blurring ==
                2.0*np.array([[4.5, -4.5], [4.5, -1.5], [4.5, 1.5], [4.5, 4.5],
                              [1.5, -4.5], [1.5, 4.5], [-1.5, -4.5], [-1.5, 4.5],
                              [-4.5, -4.5], [-4.5, -1.5], [-4.5, 1.5], [-4.5, 4.5]])).all()

    def test_padded_grid_stack(self, lens_data_stack):

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=np.full((6, 6), False),
                                                            pixel_scales=lens_data_stack.images[0].pixel_scales)

        assert (lens_data_stack.padded_grid_stacks[0].regular == padded_image_util).all()
        assert lens_data_stack.padded_grid_stacks[0].regular.image_shape == (4, 4)
        assert lens_data_stack.padded_grid_stacks[0].regular.padded_shape == (6, 6)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((6, 6), False), pixel_scales=lens_data_stack.images[0].pixel_scales,
            sub_grid_size=lens_data_stack.grid_stacks[0].sub.sub_grid_size)

        assert lens_data_stack.padded_grid_stacks[0].sub == pytest.approx(padded_sub_util, 1e-4)
        assert lens_data_stack.padded_grid_stacks[0].sub.image_shape == (4, 4)
        assert lens_data_stack.padded_grid_stacks[0].sub.padded_shape == (6, 6)
        assert (lens_data_stack.padded_grid_stacks[0].blurring == np.array([[0.0, 0.0]])).all()

        padded_image_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=np.full((6, 6), False),
                                                            pixel_scales=lens_data_stack.images[1].pixel_scales)

        assert (lens_data_stack.padded_grid_stacks[1].regular == padded_image_util).all()
        assert lens_data_stack.padded_grid_stacks[1].regular.image_shape == (4, 4)
        assert lens_data_stack.padded_grid_stacks[1].regular.padded_shape == (6, 6)

        padded_sub_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((6, 6), False), pixel_scales=lens_data_stack.images[1].pixel_scales,
            sub_grid_size=lens_data_stack.grid_stacks[1].sub.sub_grid_size)

        assert lens_data_stack.padded_grid_stacks[1].sub == pytest.approx(padded_sub_util, 1e-4)
        assert lens_data_stack.padded_grid_stacks[1].sub.image_shape == (4, 4)
        assert lens_data_stack.padded_grid_stacks[1].sub.padded_shape == (6, 6)
        assert (lens_data_stack.padded_grid_stacks[1].blurring == np.array([[0.0, 0.0]])).all()

        # Pixel scale is doubled, thus the grids of coordinates are doubled.

        assert (2.0*lens_data_stack.padded_grid_stacks[0].regular ==
                lens_data_stack.padded_grid_stacks[1].regular).all()
        assert (2.0*lens_data_stack.padded_grid_stacks[0].sub ==
                lens_data_stack.padded_grid_stacks[1].sub).all()

    def test_border(self, lens_data_stack):
        assert (lens_data_stack.borders[0] == np.array([0, 1, 2, 3])).all()
        assert (lens_data_stack.borders[1] == np.array([0, 1, 2, 3])).all()

    def test_convolvers(self, lens_data_stack):
        assert type(lens_data_stack.convolvers_image[0]) == convolution.ConvolverImage
        assert type(lens_data_stack.convolvers_mapping_matrix[0]) == inversion_convolution.ConvolverMappingMatrix
        assert type(lens_data_stack.convolvers_image[1]) == convolution.ConvolverImage
        assert type(lens_data_stack.convolvers_mapping_matrix[1]) == inversion_convolution.ConvolverMappingMatrix

    def test__constructor_inputs(self):

        psf_0 = im.PSF(np.ones((7, 7)), 1)
        ccd_0 = im.CCDData(np.ones((51, 51)), pixel_scale=3., psf=psf_0, noise_map=np.ones((51, 51)))
        mask_0 = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(51, 51), pixel_scale=1.0, invert=True)
        mask_0[26, 26] = False

        psf_1 = im.PSF(np.ones((7, 7)), 1)
        ccd_1 = im.CCDData(np.ones((51, 51)), pixel_scale=3., psf=psf_1, noise_map=np.ones((51, 51)))
        mask_1 = msk.Mask.unmasked_for_shape_and_pixel_scale(shape=(51, 51), pixel_scale=1.0, invert=True)
        mask_1[26, 26] = False

        lens_data_stack = lis.LensDataStack(ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1], sub_grid_size=8,
                                             image_psf_shape=(5, 5), mapping_matrix_psf_shape=(3, 3),
                                             positions=[np.array([[1.0, 1.0]])])

        assert lens_data_stack.sub_grid_size == 8
        assert lens_data_stack.convolvers_image[0].psf_shape == (5, 5)
        assert lens_data_stack.convolvers_mapping_matrix[0].psf_shape == (3, 3)
        assert lens_data_stack.convolvers_image[1].psf_shape == (5, 5)
        assert lens_data_stack.convolvers_mapping_matrix[1].psf_shape == (3, 3)
        assert (lens_data_stack.positions[0] == np.array([[1.0, 1.0]])).all()


@pytest.fixture(name="lens_hyper_image_stack")
def make_lens_hyper_image(ccd_0, ccd_1, mask_0, mask_1):

    return lis.LensHyperDataStack(ccd_datas=[ccd_0, ccd_1], masks=[mask_0, mask_1],
                                  hyper_model_images=[10.0*np.ones((4,4)), 20.0*np.ones((4,4))],
                                  hyper_galaxy_images_stack=[[11.0*np.ones((4,4)), 12.0*np.ones((4,4))],
                                                                  [21.0*np.ones((4,4)), 22.0*np.ones((4,4))]],
                                  hyper_minimum_values=[0.1, 0.2])


class TestLensingHyperImage(object):

    def test_attributes(self, ccd_0, ccd_1, lens_hyper_image_stack):

        assert lens_hyper_image_stack.pixel_scales[0] == ccd_0.pixel_scale

        assert (lens_hyper_image_stack.images[0] == ccd_0.image).all()
        assert (lens_hyper_image_stack.images[0] == np.ones((4,4))).all()

        assert (lens_hyper_image_stack.psfs[0] == ccd_0.psf).all()
        assert (lens_hyper_image_stack.psfs[0] == np.ones((3,3))).all()

        assert (lens_hyper_image_stack.noise_maps[0] == ccd_0.noise_map).all()
        assert (lens_hyper_image_stack.noise_maps[0] == 2.0*np.ones((4,4))).all()

        assert (lens_hyper_image_stack.hyper_model_images[0] == 10.0*np.ones((4,4))).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_stack[0][0] == 11.0*np.ones((4,4))).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_stack[0][1] == 12.0*np.ones((4,4))).all()

        assert lens_hyper_image_stack.pixel_scales[1] == ccd_1.pixel_scale

        assert (lens_hyper_image_stack.images[1] == ccd_1.image).all()
        assert (lens_hyper_image_stack.images[1] == 10.0*np.ones((4,4))).all()

        assert (lens_hyper_image_stack.psfs[1] == ccd_1.psf).all()
        assert (lens_hyper_image_stack.psfs[1] == 11.0*np.ones((3,3))).all()

        assert (lens_hyper_image_stack.noise_maps[1] == ccd_1.noise_map).all()
        assert (lens_hyper_image_stack.noise_maps[1] == 12.0*np.ones((4,4))).all()

        assert (lens_hyper_image_stack.hyper_model_images[1] == 20.0*np.ones((4,4))).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_stack[1][0] == 21.0*np.ones((4,4))).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_stack[1][1] == 22.0*np.ones((4,4))).all()
        
        assert lens_hyper_image_stack.hyper_minimum_values == [0.1, 0.2]

    def test_masking(self, lens_hyper_image_stack):

        assert (lens_hyper_image_stack.images_1d[0] == np.ones(4)).all()
        assert (lens_hyper_image_stack.noise_maps_1d[0] == 2.0*np.ones(4)).all()

        assert (lens_hyper_image_stack.hyper_model_images_1d[0] == 10.0*np.ones(4)).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_1d_stack[0][0] == 11.0*np.ones(4)).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_1d_stack[0][1] == 12.0*np.ones(4)).all()

        assert (lens_hyper_image_stack.images_1d[1] == 10.0*np.ones(4)).all()
        assert (lens_hyper_image_stack.noise_maps_1d[1] == 12.0*np.ones(4)).all()

        assert (lens_hyper_image_stack.hyper_model_images_1d[1] == 20.0*np.ones(4)).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_1d_stack[1][0] == 21.0*np.ones(4)).all()
        assert (lens_hyper_image_stack.hyper_galaxy_images_1d_stack[1][1] == 22.0*np.ones(4)).all()


from autofit.optimize import non_linear as nl
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline

def make_pipeline(pipeline_path=''):

    pipeline_name = 'pipeline_complex_source'
    pipeline_path = pipeline_path + pipeline_name

    # To begin, we need to initialize the lens's mass model. We should be able to do this by using a simple source
    # model. It won't fit the complicated structure of the source, but it'll give us a reasonable estimate of the
    # einstein radius and the other lens-mass parameters.

    # This should run fine without any prior-passes. In general, a thick, giant ring of source light is something we
    # can be confident MultiNest will fit without much issue, especially when the lens galaxy's light isn't included
    # such that the parameter space is just 12 parameters.

    phase1 = ph.LensSourcePlanePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                                     source_galaxies=dict(source=gm.GalaxyModel(light_0=lp.EllipticalSersic)),
                                     optimizer_class=nl.MultiNest, phase_name=pipeline_path + '/phase_1_simple_source')

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.5

    # Now lets add another source component, using the previous model as the initialization on the lens / source
    # parameters. We'll vary the parameters of the lens mass model and first source galaxy component during the fit.

    class X2SourcePhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens = previous_results[0].variable.lens
            self.source_galaxies.source.light_0 = previous_results[0].variable.source.light_0

    # You'll notice I've stop writing 'phase_1_results = previous_results[0]' - we know how
    # the previous results are structured now so lets not clutter our code!

    phase2 = X2SourcePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                           source_galaxies=dict(source=gm.GalaxyModel(light_0=lp.EllipticalExponential,
                                                                       light_1=lp.EllipticalSersic)),
                           optimizer_class=nl.MultiNest, phase_name=pipeline_path + '/phase_2_x2_source')

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.5

    # Now lets do the same again, but with 3 source galaxy components.

    class X3SourcePhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens = previous_results[1].variable.lens
            self.source_galaxies.source.light_0 = previous_results[1].variable.source.light_0
            self.source_galaxies.source.light_1 = previous_results[1].variable.source.light_1

    phase3 = X3SourcePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                           source_galaxies=dict(source=gm.GalaxyModel(light_0=lp.EllipticalExponential,
                                                                      light_1=lp.EllipticalSersic,
                                                                      light_2=lp.EllipticalSersic)),
                           optimizer_class=nl.MultiNest, phase_name=pipeline_path + '/phase_3_x3_source')

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.5

    # And one more for luck!

    class X4SourcePhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens = previous_results[2].variable.lens
            self.source_galaxies.source.light_0 = previous_results[2].variable.source.light_0
            self.source_galaxies.source.light_1 = previous_results[2].variable.source.light_1
            self.source_galaxies.source.light_2 = previous_results[2].variable.source.light_2

    phase4 = X4SourcePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                           source_galaxies=dict(source=gm.GalaxyModel(light_0=lp.EllipticalExponential,
                                                                      light_1=lp.EllipticalSersic,
                                                                      light_2=lp.EllipticalSersic,
                                                                      light_3=lp.EllipticalSersic)),
                           optimizer_class=nl.MultiNest, phase_name=pipeline_path + '/phase_4_x4_source')

    phase4.optimizer.const_efficiency_mode = True
    phase4.optimizer.n_live_points = 50
    phase4.optimizer.sampling_efficiency = 0.5

    return pipeline.PipelineImaging(pipeline_path, phase1, phase2, phase3, phase4)


# In the previous 3  pipelines, we passed priors using the 'variable' attribute of the previous results. However, its
# not yet clear how these priors are passed. Do they use a UniformPrior or GaussianPrior? What are the limits / mean /
# width of these priors? Can I change this behaviour?

# Lets say I link two parameters in pass priors (don't run this code its just a demo)

def pass_priors(self, previous_results):

    self.lens_galaxies.galaxy_name.profile_name.parameter_name = \
        previous_results[0].variable.galaxy_name.profile_name.parameter_name

# By invoking the 'variable' attribute, the passing of priors behaves following 3 rules:

# 1) The 'self.lens_galaxies.galaxy_name.profile_name.parameter_name' parameter will use a GaussianPrior as its prior.

#    A GaussianPrior is ideal, as the 1D pdf results we compute at the end of a phase are easily summarized as a
#    Gaussian.

# 2) The mean of the GaussianPrior is the best-fit value of
#    'previous_results[0].variable.galaxy_name.profile_name.parameter_name'.

#    This means that MultiNest specifically starts by searching the region of non-linear parameter space that
#    corresponded to highest likelihood solutions in the previous phase. Thus, we're setting our priors to look in the
#    correct regions of parameter space.

# 3) The sigma of the Gaussian will use either: (i) the 1D error on the previous result's parameter or; (ii) the value
#    specified in the appropriate 'config/priors/width/profile.ini' config file (check these files out now).

#    The idea here is simple. We want a value of sigma that gives a GaussianPrior wide enough to search a broad region
#    of parameter space, so that the lens model can change if a better solution is nearby. However, we want it to be
#    narrow enough that we don't search too much of parameter space, as this will be slow or risk leading us into an
#    incorrect solution! A natural choice is the errors of the parameter from the previous phase.

#    Unfortunately, this doesn't always work. Lens modeling is prone to an effect called 'over-fitting' where we
#    underestimate the errors on our lens model parameters. This is especially true when we take the shortcuts we're
#    used to in early phases - aggresive masking, reduced data, simplified lens models, constant efficiency mode, etc.

#    Therefore, the priors/widths file is our fallback. If the error on a parameter is suspiciously small, we instead
#    use the value specified in the widths file. These values are chosen based on our experience as being a good
#    balance broadly sampling parameter space but not being so narrow important solutions are missed. There are two
#    ways a value is specified using the priors/width file:

#   1) Absolute value - 'a' - In this case, the error assumed on the parameter is the value given in the config file.
#      For example, for the width on centre_0 of a light profile, the config file reads centre_0 = a, 0.05. This means
#      if the error on the parameter centre_0 was less than 0.05 in the previous phase, the sigma of its GaussianPrior
#      in this phase will be 0.05.

#   2) Relative value - 'r' - In this case, the error assumed on the parameter is the % of the value of the best-fit
#      value given in the config file. For example, if the intensity estimated in the previous phase was 2.0, and the
#      relative error in the config file is specified as intensity = r, 0.5, then the sigma of the GaussianPrior will be
#      50% of this best-fit value, i.e. sigma = 0.5 * 2.0 = 1.0.

# We use absolute and relative values for different parameters, depending on their properties. For example, using the
# relative value of a parameter like the profile centre makes no sense. If our lens galaxy is centred at (0.0, 0.0),
# the relative error will always be tiny and thus poorly defined. Therefore, the default configs in PyAutoLens use
# absolute errors on the centre.

# However, there are parameters where using an absolute value does not make sense. Intensity is a good example of this.
# The intensity of an image depends on its units, S/N, galaxy brightness, profile definition, etc. There is no single
# absolute value that one can use to generically link the intensity of any two proflies. Thus, it makes more sense
# to link them using the relative value from a previous phase.

### EXAMPLE ##

# Lets go through an example using a real parameter. Lets say in phase 1 we fit the lens galaxy's light with an
# elliptical Sersic profile, and we estimate that its sersic index is equal to 4.0 +- 2.0. To pass this as a prior to
# phase 2, we would write:

def pass_priors(self, previous_results):

    self.lens_galaxies.lens.light.sersic_index = previous_results[0].variable.lens.light.sersic

# The prior on the lens galaxy's sersic light profile would thus be a GaussianPrior in phase 2, with mean=4.0 and
# sigma=2.0.

# If the error on the Sersic index in phase 1 had been really small, lets say, 0.01, we would use the value of the
# Sersic index width in the priors/width config file to set sigma instead. In this case, the prior config file specifies
# that we use an absolute value of 0.8 to link this prior. Thus, the GaussianPrior in phase 2 would have a mean=4.0 and
# sigma=0.8.

# If the prior config file had specified that we use an relative value of 0.8, the GaussianPrior in phase 2 would have
# a mean=4.0 and sigma=3.2.

# And with that, we're done. Linking priors is a bit of an art form, but one that tends to work really well. Its true
# to say that things can go wrong - maybe we 'trim' out the solution we're looking for, or underestimate our errors a
# bit due to making our priors too narrow. However, in general, things are okay, the point is that you should test
# pipelines with different settings, and settle on a setup that appears to be give consistent results but the faster
# run times.


import logging

from autolens import decorator_util
import numpy as np

from autolens.data.array.util import mask_util

logger = logging.getLogger(__name__)
logger.level = logging.DEBUG

@decorator_util.jit()
def map_1d_indexes_to_2d_indexes_for_shape(indexes_1d, shape):

    indexes_2d = np.zeros((indexes_1d.shape[0], 2))

    for i, index_1d in enumerate(indexes_1d):

        indexes_2d[i,0] = int(index_1d / shape[1])
        indexes_2d[i,1] = int(index_1d % shape[1])

    return indexes_2d

@decorator_util.jit()
def map_2d_indexes_to_1d_indexes_for_shape(indexes_2d, shape):

    indexes_1d = np.zeros(indexes_2d.shape[0])

    for i in range(indexes_2d.shape[0]):

        indexes_1d[i] = int((indexes_2d[i,0])*shape[1] + indexes_2d[i,1])

    return indexes_1d

@decorator_util.jit()
def sub_to_regular_from_mask(mask, sub_grid_size):
    """Compute a 1D array that maps every unmasked pixel's sub-pixel to its corresponding 1d datas_-pixel.

    For howtolens, if sub-pixel 8 is in datas_-pixel 1, sub_to_regular[7] = 1."""

    total_sub_pixels = mask_util.total_sub_pixels_from_mask_and_sub_grid_size(mask, sub_grid_size)

    sub_to_regular = np.zeros(shape=total_sub_pixels)
    regular_index = 0
    sub_index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                for y1 in range(sub_grid_size):
                    for x1 in range(sub_grid_size):
                        sub_to_regular[sub_index] = regular_index
                        sub_index += 1

                regular_index += 1

    return sub_to_regular

@decorator_util.jit()
def map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d):
    """For a given 2D array and masks, mappers all unmasked pixels to a 1D array."""

    total_image_pixels = mask_util.total_regular_pixels_from_mask(mask)

    array_1d = np.zeros(shape=total_image_pixels)
    index = 0

    for y in range(mask.shape[0]):
        for x in range(mask.shape[1]):
            if not mask[y, x]:
                array_1d[index] = array_2d[y, x]
                index += 1

    return array_1d

@decorator_util.jit()
def map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(array_1d, shape, one_to_two):
    """For a masked 1D array, mappers it to a 2D array using the mappings from 1D to 2D (one_to_two)."""

    array_2d = np.zeros(shape)

    for index in range(len(one_to_two)):
        array_2d[one_to_two[index, 0], one_to_two[index, 1]] = array_1d[index]

    return array_2d

@decorator_util.jit()
def map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d, shape):
    """Map a 1D array where every value in its original 2D array was unmasked, to this original 2D array."""

    array_2d = np.zeros(shape)

    index = 0
    for y in range(shape[0]):
        for x in range(shape[1]):
            array_2d[y, x] = array_1d[index]
            index += 1

    return array_2d

@decorator_util.jit()
def sparse_to_unmasked_sparse_from_mask_and_pixel_centres(total_sparse_pixels, mask, unmasked_sparse_grid_pixel_centres):
    """Determine the mapping between every masked pixelization-grid pixel and pixelization-grid pixel. This is
    performed by checking whether each pixelization-grid pixel is within the regular-masks, and mapping the indexes.

    Parameters
    -----------
    total_sparse_pixels : int
        The total number of pixels in the pixelization grid which fall within the regular-masks.
    mask : ccd.masks.Mask
        The regular-masks within which pixelization pixels must be inside
    unmasked_sparse_grid_pixel_centres : ndarray
        The centres of the unmasked pixelization grid pixels.
    """

    pix_to_full_pix = np.zeros(total_sparse_pixels)

    pixel_index = 0

    for full_pixel_index in range(unmasked_sparse_grid_pixel_centres.shape[0]):

        y = unmasked_sparse_grid_pixel_centres[full_pixel_index, 0]
        x = unmasked_sparse_grid_pixel_centres[full_pixel_index, 1]

        if not mask[y, x]:

            pix_to_full_pix[pixel_index] = full_pixel_index
            pixel_index += 1

    return pix_to_full_pix

@decorator_util.jit()
def unmasked_sparse_to_sparse_from_mask_and_pixel_centres(mask, unmasked_sparse_grid_pixel_centres, total_sparse_pixels):
    """Determine the mapping between every pixelization-grid pixel and masked pixelization-grid pixel. This is
    performed by checking whether each pixelization-grid pixel is within the regular-masks, and mapping the indexes.

    Pixelization pixels are paired with the next masked pixel index. This may mean that a pixel is not paired with a
    pixel near it, if the next pixel is on the next row of the grid. This is not a problem, as it is only
    unmasked pixels that are referened when computing image_to_pix, which is what this array is used for.

    Parameters
    -----------
    total_pix_pixels : int
        The total number of pixels in the pixelization grid which fall within the regular-masks.
    mask : ccd.masks.Mask
        The regular-masks within which pixelization pixels must be inside
    unmasked_sparse_grid_pixel_centres : ndarray
        The centres of the unmasked pixelization grid pixels.
    """

    total_unmasked_sparse_pixels = unmasked_sparse_grid_pixel_centres.shape[0]

    unmasked_sparse_to_sparse = np.zeros(total_unmasked_sparse_pixels)
    pixel_index = 0

    for unmasked_sparse_pixel_index in range(total_unmasked_sparse_pixels):

        y = unmasked_sparse_grid_pixel_centres[unmasked_sparse_pixel_index, 0]
        x = unmasked_sparse_grid_pixel_centres[unmasked_sparse_pixel_index, 1]

        unmasked_sparse_to_sparse[unmasked_sparse_pixel_index] = pixel_index

        if not mask[y, x]:
            if pixel_index < total_sparse_pixels-1:
                pixel_index += 1

    return unmasked_sparse_to_sparse

@decorator_util.jit()
def regular_to_sparse_from_sparse_mappings(regular_to_unmasked_sparse, unmasked_sparse_to_sparse):
    """Using the mapping between the regular-grid and unmasked pixelization grid, compute the mapping between each regular
    pixel and the masked pixelization grid.

    Parameters
    -----------
    regular_to_unmasked_sparse : ndarray
        The index mapping between every regular-pixel and masked pixelization pixel.
    unmasked_sparse_to_sparse : ndarray
        The index mapping between every masked pixelization pixel and unmasked pixelization pixel.
    """
    total_regular_pixels = regular_to_unmasked_sparse.shape[0]

    regular_to_sparse = np.zeros(total_regular_pixels)

    for regular_index in range(total_regular_pixels):

        regular_to_sparse[regular_index] = unmasked_sparse_to_sparse[regular_to_unmasked_sparse[regular_index]]

    return regular_to_sparse

@decorator_util.jit()
def sparse_grid_from_unmasked_sparse_grid(unmasked_sparse_grid, sparse_to_unmasked_sparse):
    """Use the central arc-second coordinate of every unmasked pixelization grid's pixels and mapping between each
    pixelization pixel and unmasked pixelization pixel to compute the central arc-second coordinate of every masked
    pixelization grid pixel.

    Parameters
    -----------
    unmasked_sparse_grid : ndarray
        The (y,x) arc-second centre of every unmasked pixelization grid pixel.
    sparse_to_unmasked_sparse : ndarray
        The index mapping between every pixelization pixel and masked pixelization pixel.
    """
    total_pix_pixels = sparse_to_unmasked_sparse.shape[0]

    pix_grid = np.zeros((total_pix_pixels, 2))

    for pixel_index in range(total_pix_pixels):

        pix_grid[pixel_index, 0] = unmasked_sparse_grid[sparse_to_unmasked_sparse[pixel_index], 0]
        pix_grid[pixel_index, 1] = unmasked_sparse_grid[sparse_to_unmasked_sparse[pixel_index], 1]

    return pix_grid


from autofit import conf
from autolens.data import ccd
from autolens.data.array import mask as msk
from autolens.data.plotters import ccd_plotters

import os

path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))
conf.instance = conf.Config(config_path=path+'config', output_path=path+'output')

# lens_name = 'slacs0216-0813' # Incorrect
lens_name = 'slacs0252+0039' # Works
lens_name = 'slacs0737+3216' # Works
lens_name = 'slacs0912+0029' # Works
lens_name = 'slacs0959+4410' # Bad fit phase 5 due to source demag thing
lens_name = 'slacs0959+4416' # Bad Fit because its the hard to fit lens
# lens_name = 'slacs1011+0143' # x2 lenses
# lens_name = 'slacs1205+4910' # Bad fit because foreground galaxy is fitted
# lens_name = 'slacs1250+0523' # Bad fit due to poor parametric source.
# lens_name = 'slacs1402+6321' # Bad fit and i cant even see the source :/
# lens_name = 'slacs1420+6019' # Bad fit phase 5 due to complex source.
# lens_name = 'slacs1430+4105' # Works
# lens_name = 'slacs1627+0053' # Works, but weird low res source
# lens_name = 'slacs1630+4520'
lens_name = 'slacs2238-0754'
# lens_name = 'slacs2300+0022'
# lens_name = 'slacs2303+1422'

ccd_data = ccd.load_ccd_data_from_fits(image_path=path + '/data/slacs/' + lens_name + '/F814W_image.fits',
                                       psf_path=path+'/data/slacs/'+lens_name+'/F814W_psf.fits',
                                       noise_map_path=path+'/data/slacs/'+lens_name+'/F814W_noise_map.fits',
                                       pixel_scale=0.03, resized_ccd_shape=(301, 301), resized_psf_shape=(21,21))

mask = msk.Mask.circular(shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0)

ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data, mask=mask)

from workspace_jam.pipelines import lens_light_and_source_inversion

pipeline = lens_light_and_source_inversion.make_pipeline(pipeline_path='slacs/' + lens_name + '/')

pipeline.run(data=ccd_data)


from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autofit.tools import phase as autofit_ph
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# In this pipeline, we'll perform a subhalo analysis which fits a source galaxy using a light profile
# followed by an inversion, where the lens galaxy's light is not present in the image. Two grid-searches will then be
# performed, to assess our sensitivity to subhalos and to attempt to detect subhalos in the observed image. The
# pipeline comprises four phases:

# Phase 1) Fit the lens galaxy's mass (SIE) and source galaxy's light using a single Sersic light profile.

# Phase 2) Fit the lens galaxy's mass (SIE) and source galaxy's light using an inversion, initializing the priors
#          of the lens the results of Phase 1.

# Phase 3) Perform the sensitivity analysis, using an ordinary grid search over subhalo (y,x) coordinates and mass.

# Phase 4) Perform the subhalo detection analysis, using a Multinest grid search oer subhalo (y,x) coordinate cells,
#          where each phase optimizes the subhalo position in these cells as well as its mass and concentration.

def make_pipeline(pipeline_path=''):

    pipeline_name = 'pipeline_subhalo_no_lens_light'
    pipeline_path = pipeline_path + pipeline_name

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function_annular(image):
        return msk.Mask.circular_annular(shape=image.shape, pixel_scale=image.pixel_scale,
                                         inner_radius_arcsec=0.2, outer_radius_arcsec=3.3)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    class LensSourcePhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens.mass.centre_0 = prior.GaussianPrior(mean=0.0, sigma=0.1)
            self.lens_galaxies.lens.mass.centre_1 = prior.GaussianPrior(mean=0.0, sigma=0.1)

    phase1 = LensSourcePhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                               source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalSersic)),
                               mask_function=mask_function_annular, optimizer_class=nl.MultiNest,
                               phase_name=pipeline_path + '/phase_1_source')

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we fit the lens's mass and source galaxy using an inversion, where we:

    # 1) Initialize the priors on the lens galaxy using the results of phase 1.
    # 2) Assume default priors for all source inversion parameters.

    class InversionPhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens.mass = previous_results[0].variable.lens.mass

    phase2 = InversionPhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal)),
                            source_galaxies=dict(source=gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                                      regularization=reg.Constant)),
                            optimizer_class=nl.MultiNest, mask_function=mask_function_annular,
                            phase_name=pipeline_path + '/phase_2_inversion')

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.5

    ### Phase 3 ###

    # In phase 3, we perform the sensitivity analysis of our lens, using a grid search of subhalo (y,x) coordinates and
    # mass, where:

    # 1) The lens model and source-pixelization parameters are held fixed to the best-fit values from phase 2.

    class GridPhase(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):

            self.lens_galaxies.lens.mass = previous_results[1].constant.lens.mass
            self.source_galaxies.source.pixelization = previous_results[1].constant.source.pixelization
            self.source_galaxies.source.regularization = previous_results[1].constant.source.regularization

            self.lens_galaxies.subhalo.mass.centre_0 = prior.UniformPrior(lower_limit=-2.0, upper_limit=2.0)
            self.lens_galaxies.subhalo.mass.centre_1 = prior.UniformPrior(lower_limit=-2.0, upper_limit=2.0)
            self.lens_galaxies.subhalo.mass.kappa_s = prior.UniformPrior(lower_limit=0.0, upper_limit=1.0)
            self.lens_galaxies.subhalo.mass.scale_radius = 5.0

    phase3 = GridPhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal),
                                          subhalo=gm.GalaxyModel(mass=mp.SphericalNFW)),
                       source_galaxies=dict(source=gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                                  regularization=reg.Constant)),
                       optimizer_class=nl.GridSearch, mask_function=mask_function_annular,
                       phase_name=pipeline_path + '/phase_3_sensitivity')

    ### Phase 4 ###

    # In phase 4, we attempt to detect subhalos, by performing a NxN grid search of MultiNest searches, where:

    # 1) The lens model and source-pixelization parameters are held fixed to the best-fit values from phase 2.
    # 2) Each grid search varies the subhalo (y,x) coordinates and mass as free parameters.
    # 3) The priors on these (y,x) coordinates are UniformPriors, with limits corresponding to the grid-cells.

    class GridPhase(autofit_ph.as_grid_search(ph.LensSourcePlanePhase)):

        @property
        def grid_priors(self):
            return [self.variable.subhalo.mass.centre_0, self.variable.subhalo.mass.centre_1]

        def pass_priors(self, previous_results):

            self.lens_galaxies.subhalo.mass.centre_0 = prior.UniformPrior(lower_limit=-2.0, upper_limit=2.0)
            self.lens_galaxies.subhalo.mass.centre_1 = prior.UniformPrior(lower_limit=-2.0, upper_limit=2.0)
            self.lens_galaxies.subhalo.mass.kappa_s = prior.UniformPrior(lower_limit=0.0, upper_limit=0.2)

            self.lens_galaxies.lens.mass = previous_results[1].constant.lens.mass
            self.source_galaxies.source.pixelization = previous_results[1].constant.source.pixelization
            self.source_galaxies.source.regularization = previous_results[1].constant.source.regularization


    phase4 = GridPhase(lens_galaxies=dict(lens=gm.GalaxyModel(mass=mp.EllipticalIsothermal),
                                          subhalo=gm.GalaxyModel(mass=mp.SphericalNFW)),
                       source_galaxies=dict(source=gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                                  regularization=reg.Constant)),
                       number_of_steps=6, optimizer_class=nl.MultiNest,
                       phase_name=pipeline_path + '/phase_4_subhalo_search')

    return pipeline.PipelineImaging(pipeline_path, phase1, phase2, phase3, phase4)


import numpy as np
import scipy.spatial
import sklearn.cluster

from autolens import exc
from autolens.data.array import grids, scaled_array
from autolens.model.inversion import mappers
from autolens.model.inversion.util import pixelization_util


def setup_image_plane_pixelization_grid_from_galaxies_and_grid_stack(galaxies, grid_stack):
    """An image-plane pixelization is one where its pixel centres are computed by tracing a sparse grid of pixels from \
    the image's regular grid to other planes (e.g. the source-plane).

    Provided a galaxy has an image-plane pixelization, this function returns a new *GridStack* instance where the \
    image-plane pixelization's sparse grid is added to it as an attibute.

    Thus, when the *GridStack* are are passed to the *ray_tracing* module this sparse grid is also traced and the \
    traced coordinates represent the centre of each pixelization pixel.

    Parameters
    -----------
    galaxies : [model.galaxy.galaxy.Galaxy]
        A list of galaxies, which may contain pixelizations and an *ImagePlanePixelization*.
    grid_stacks : image.array.grid_stacks.GridStack
        The collection of grid_stacks (regular, sub, etc.) which the image-plane pixelization grid (referred to as pix) \
        may be added to.
    """
    if not isinstance(grid_stack.regular, grids.PaddedRegularGrid):
        for galaxy in galaxies:
            if hasattr(galaxy, 'pixelization'):
                if isinstance(galaxy.pixelization, ImagePlanePixelization):

                    image_plane_pix_grid = galaxy.pixelization.image_plane_pix_grid_from_regular_grid(
                        regular_grid=grid_stack.regular)
                    return grid_stack.grid_stack_with_pix_grid_added(pix_grid=image_plane_pix_grid.sparse_grid,
                                                                     regular_to_nearest_pix=image_plane_pix_grid.regular_to_sparse)

    return grid_stack


class ImagePlanePixelization(object):

    def __init__(self, shape):
        """An image-plane pixelization is one where its pixel centres are computed by tracing a sparse grid of pixels \
        from the image's regular grid to other planes (e.g. the source-plane).

        The traced coordinates of this sparse grid represent each centre of a pixelization pixel.

        See *grid_stacks.SparseToRegularGrid* for details on how this grid is calculated.

        Parameters
        -----------
        shape : (float, float) or (int, int)
            The shape of the image-plane pixelizaton grid in pixels (floats are converted to integers). The grid is \
            laid over the masked image such that it spans the most outer pixels of the masks.
        """
        self.shape = (int(shape[0]), int(shape[1]))

    def image_plane_pix_grid_from_regular_grid(self, regular_grid):
        """Calculate the image-plane pixelization from a regular-grid of coordinates (and its mask).

        See *grid_stacks.SparseToRegularGrid* for details on how this grid is calculated.

        Parameters
        -----------
        regular_grid : grids.RegularGrid
            The grid of (y,x) arc-second coordinates at the centre of every image value (e.g. image-pixels).
        """
        imagepixel_scale = regular_grid.mask.pixel_scale
        pixel_scales = ((regular_grid.masked_shape_arcsec[0] + imagepixel_scale) / self.shape[0],
                        (regular_grid.masked_shape_arcsec[1] + imagepixel_scale) / self.shape[1])
        return grids.SparseToRegularGrid(unmasked_sparse_grid_shape=self.shape, pixel_scales=pixel_scales,
                                         regular_grid=regular_grid, origin=regular_grid.mask.centre)


class Pixelization(object):

    def __init__(self):
        """ Abstract base class for a pixelization, which discretizes grid_stack of (y,x) coordinates into pixels.
        """

    def mapper_from_grid_stack_and_border(self, grid_stack, border):
        raise NotImplementedError("pixelization_mapper_from_grids_and_borders should be overridden")

    def __str__(self):
        return "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()])

    def __repr__(self):
        return "{}\n{}".format(self.__class__.__name__, str(self))


class Rectangular(Pixelization):

    def __init__(self, shape=(3, 3)):
        """A rectangular pixelization, where pixels are defined on a Cartesian and uniform grid of shape \ 
        (rows, columns).

        Like arrays, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (y_pixels, x_pixel)
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular pixelization must be at least dimensions 3x3')

        self.shape = (int(shape[0]), int(shape[1]))
        self.pixels = self.shape[0] * self.shape[1]
        super(Rectangular, self).__init__()

    class Geometry(scaled_array.RectangularArrayGeometry):

        def __init__(self, shape, pixel_scales, origin, pixel_neighbors, pixel_neighbors_size):
            """The geometry of a rectangular grid.

            This is used to map grid_stack of (y,x) arc-second coordinates to the pixels on the rectangular grid.

            Parameters
            -----------
            shape : (int, int)
                The dimensions of the rectangular grid of pixels (y_pixels, x_pixel)
            pixel_scales : (float, float)
                The pixel-to-arcsecond scale of a pixel in the y and x directions.
            origin : (float, float)
                The arc-second origin of the rectangular pixelization's coordinate system.
            pixel_neighbors : ndarray
                An array of length (y_pixels*x_pixels) which provides the index of all neighbors of every pixel in \
                the rectangular grid (entries of -1 correspond to no neighbor).
            pixel_neighbors_size : ndarrayy
                An array of length (y_pixels*x_pixels) which gives the number of neighbors of every pixel in the \
                rectangular grid.
            """
            self.shape = shape
            self.pixel_scales = pixel_scales
            self.origin = origin
            self.pixel_neighbors = pixel_neighbors.astype('int')
            self.pixel_neighbors_size = pixel_neighbors_size.astype('int')

        @property
        def pixel_centres(self):
            """The centre of every pixel in the rectangular pixelization."""
            return self.grid_1d

    def geometry_from_grid(self, grid, buffer=1e-8):
        """Determine the geometry of the rectangular grid, by overlaying it over a grid of coordinates such that its \
         outer-most pixels align with the grid's outer most coordinates plus a small buffer.

        Parameters
        -----------
        grid : ndarray
            The (y,x) grid of coordinates over which the rectangular pixelization is placed to determine its geometry.
        buffer : float
            The size the pixelization is buffered relative to the grid.
        """
        y_min = np.min(grid[:, 0]) - buffer
        y_max = np.max(grid[:, 0]) + buffer
        x_min = np.min(grid[:, 1]) - buffer
        x_max = np.max(grid[:, 1]) + buffer
        pixel_scales = (float((y_max - y_min) / self.shape[0]), float((x_max - x_min) / self.shape[1]))
        origin = ((y_max + y_min) / 2.0, (x_max + x_min) / 2.0)
        pixel_neighbors, pixel_neighbors_size = self.neighbors_from_pixelization()
        return self.Geometry(shape=self.shape, pixel_scales=pixel_scales, origin=origin,
                             pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

    def neighbors_from_pixelization(self):
        return pixelization_util.rectangular_neighbors_from_shape(shape=self.shape)

    def mapper_from_grid_stack_and_border(self, grid_stack, border):
        """Setup a rectangular mapper from a rectangular pixelization, as follows:

        1) If a border is supplied, relocate all of the grid-stack's regular and sub grid pixels beyond the border.
        2) Determine the rectangular pixelization's geometry, by laying the pixelization over the sub-grid.
        3) Setup the rectangular mapper from the relocated grid-stack and rectangular pixelization.

        Parameters
        ----------
        grid_stack : grids.GridStack
            A stack of grid describing the observed image's pixel coordinates (e.g. an image-grid, sub-grid, etc.).
        border : grids.RegularGridBorder
            The border of the grid-stack's regular-grid.
        """

        if border is not None:
            relocated_grid_stack = border.relocated_grid_stack_from_grid_stack(grid_stack)
        else:
            relocated_grid_stack = grid_stack

        geometry = self.geometry_from_grid(grid=relocated_grid_stack.sub)

        return mappers.RectangularMapper(pixels=self.pixels, grid_stack=relocated_grid_stack, border=border,
                                         shape=self.shape, geometry=geometry)


class Voronoi(Pixelization):

    def __init__(self):
        """Abstract base class for a Voronoi pixelization, which represents pixels as an irregular grid of Voronoi \
         cells which can form any shape, size or tesselation.

         The grid-stack's coordinates are paired to Voronoi pixels as the nearest-neighbors of the Voronoi \
        pixel-centers.
         """
        super(Voronoi, self).__init__()

    class Geometry(scaled_array.ArrayGeometry):

        def __init__(self, shape_arc_seconds, pixel_centres, origin, pixel_neighbors, pixel_neighbors_size):
            """The geometry of a Voronoi pixelization.

            Parameters
            -----------
            shape_arc_seconds : (float, float)
                The dimensions of the Voronoi grid ni arc-second (y_arcseconds, x_arcseconds)
            pixel_centres : ndarray
                The (y,x) centre of every Voronoi pixel in arc-seconds.
            origin : (float, float)
                The arc-second origin of the Voronoi pixelization's coordinate system.
            pixel_neighbors : ndarray
                An array of length (voronoi_pixels) which provides the index of all neighbors of every pixel in \
                the Voronoi grid (entries of -1 correspond to no neighbor).
            pixel_neighbors_size : ndarrayy
                An array of length (voronoi_pixels) which gives the number of neighbors of every pixel in the \
                Voronoi grid.
            """
            self.shape_arc_sec = shape_arc_seconds
            self.pixel_centres = pixel_centres
            self.origin = origin
            self.pixel_neighbors = pixel_neighbors.astype('int')
            self.pixel_neighbors_size = pixel_neighbors_size.astype('int')


    def geometry_from_grid(self, grid, pixel_centres, pixel_neighbors, pixel_neighbors_size, buffer=1e-8):
        """Determine the geometry of the Voronoi pixelization, by alligning it with the outer-most coordinates on a \
        grid plus a small buffer.

        Parameters
        -----------
        grid : ndarray
            The (y,x) grid of coordinates which determine the Voronoi pixelization's geometry.
        pixel_centres : ndarray
            The (y,x) centre of every Voronoi pixel in arc-seconds.
        origin : (float, float)
            The arc-second origin of the Voronoi pixelization's coordinate system.
        pixel_neighbors : ndarray
            An array of length (voronoi_pixels) which provides the index of all neighbors of every pixel in \
            the Voronoi grid (entries of -1 correspond to no neighbor).
        pixel_neighbors_size : ndarrayy
            An array of length (voronoi_pixels) which gives the number of neighbors of every pixel in the \
            Voronoi grid.
        """
        y_min = np.min(grid[:, 0]) - buffer
        y_max = np.max(grid[:, 0]) + buffer
        x_min = np.min(grid[:, 1]) - buffer
        x_max = np.max(grid[:, 1]) + buffer
        shape_arc_seconds = (y_max - y_min, x_max - x_min)
        origin = ((y_max + y_min) / 2.0, (x_max + x_min) / 2.0)
        return self.Geometry(shape_arc_seconds=shape_arc_seconds, pixel_centres=pixel_centres, origin=origin,
                             pixel_neighbors=pixel_neighbors, pixel_neighbors_size=pixel_neighbors_size)

    @staticmethod
    def voronoi_from_pixel_centers(pixel_centers):
        """Compute the Voronoi grid of the pixelization, using the pixel centers.

        Parameters
        ----------
        pixel_centers : ndarray
            The (y,x) centre of every Voronoi pixel.
        """
        return scipy.spatial.Voronoi(np.asarray([pixel_centers[:, 1], pixel_centers[:, 0]]).T,
                                     qhull_options='Qbb Qc Qx Qm')


    def neighbors_from_pixelization(self, pixels, ridge_points):
        """Compute the neighbors of every Voronoi pixel as an ndarray of the pixel index's each pixel shares a \
        vertex with.

        The ridge points of the Voronoi grid are used to derive this.

        Parameters
        ----------
        ridge_points : scipy.spatial.Voronoi.ridge_points
            Each Voronoi-ridge (two indexes representing a pixel mapping_matrix).
        """
        return pixelization_util.voronoi_neighbors_from_pixels_and_ridge_points(pixels=pixels,
                                                                                ridge_points=np.asarray(ridge_points))


class AdaptiveMagnification(Voronoi, ImagePlanePixelization):

    def __init__(self, shape=(3, 3)):
        """A pixelization which adapts to the magnification pattern of a lens's mass model and uses a Voronoi \
        pixelization to discretize the grid into pixels.

        Parameters
        ----------
        shape : (int, int)
            The shape of the unmasked sparse-grid which is laid over the masked image, in order to derive the \
            adaptive-magnification pixelization (see *ImagePlanePixelization*)
        """
        super(AdaptiveMagnification, self).__init__()
        ImagePlanePixelization.__init__(self=self, shape=shape)

    def mapper_from_grid_stack_and_border(self, grid_stack, border):
        """Setup a Voronoi mapper from an adaptive-magnification pixelization, as follows:

        1) (before this routine is called), setup the 'pix' grid as part of the grid-stack, which corresponds to a \
           sparse set of pixels in the image-plane which are traced to form the pixel centres.
        2) If a border is supplied, relocate all of the grid-stack's regular, sub and pix grid pixels beyond the border.
        3) Determine the adaptive-magnification pixelization's pixel centres, by extracting them from the relocated \
           pix grid.
        4) Use these pixelization centres to setup the Voronoi pixelization.
        5) Determine the neighbors of every Voronoi cell in the Voronoi pixelization.
        6) Setup the geometry of the pixelizatioon using the relocated sub-grid and Voronoi pixelization.
        7) Setup a Voronoi mapper from all of the above quantities.

        Parameters
        ----------
        grid_stack : grids.GridStack
            A collection of grid describing the observed image's pixel coordinates (includes an image and sub grid).
        border : grids.RegularGridBorder
            The borders of the grid_stacks (defined by their image-plane masks).
        """

        if border is not None:
            relocated_grids = border.relocated_grid_stack_from_grid_stack(grid_stack)
        else:
            relocated_grids = grid_stack

        pixel_centres = relocated_grids.pix
        pixels = pixel_centres.shape[0]

        voronoi = self.voronoi_from_pixel_centers(pixel_centres)

        pixel_neighbors, pixel_neighbors_size = self.neighbors_from_pixelization(pixels=pixels,
                                                                                 ridge_points=voronoi.ridge_points)
        geometry = self.geometry_from_grid(grid=relocated_grids.sub, pixel_centres=pixel_centres,
                                           pixel_neighbors=pixel_neighbors,
                                           pixel_neighbors_size=pixel_neighbors_size)

        return mappers.VoronoiMapper(pixels=pixels, grid_stack=relocated_grids, border=border,
                                     voronoi=voronoi, geometry=geometry)